# -*- coding: utf-8 -*-

import os
import re
import sys
import traceback



########################################
# Globals
########################################

mfa_serial_format = r'^arn:aws:iam::\d+:mfa/[a-zA-Z0-9\+=,.@_-]+$'
re_mfa_serial_format = re.compile(mfa_serial_format)



########################################
# Print configuration functions
########################################

def configPrintException(enable):
    """
    Configure whether full stacktraces should be dumped in the console output

    :param enable:

    :return:
    """
    global verbose_exceptions
    verbose_exceptions = enable



########################################
# Print functions
########################################

def printDebug(msg):
    if verbose_exceptions:
        printGeneric(sys.stdout, msg)


def printError(msg, newLine = True):
    printGeneric(sys.stderr, msg, newLine)


def printException(e):
    global verbose_exceptions
    if verbose_exceptions:
        printError(str(traceback.format_exc()))
    else:
        printError(str(e))


def printGeneric(out, msg, newLine = True):
    out.write(msg)
    if newLine == True:
        out.write('\n')
    out.flush()


def printInfo(msg, newLine = True ):
    printGeneric(sys.stdout, msg, newLine)



########################################
# Prompt functions
########################################

def prompt(test_input = None):
    """
    Prompt function that works for Python2 and Python3

    :param test_input:                  Value to be returned when testing

    :return:                            Value typed by user (or passed in argument when testing)
    """
    if test_input != None:
        if type(test_input) == list and len(test_input):
            choice = test_input.pop(0)
        elif type(test_input) == list:
            choice = ''
        else:
            choice = test_input
    else:
        # Coverage: 4 missed statements
        try:
            choice = raw_input()
        except:
            choice = input()
    return choice


def prompt_4_mfa_code(activate = False, input = None):
    """
    Prompt for an MFA code

    :param activate:                    Set to true when prompting for the 2nd code when activating a new MFA device
    :param input:                       Used for unit testing

    :return:                            The MFA code
    """
    while True:
        if activate:
            prompt_string = 'Enter the next value: '
        else:
            prompt_string = 'Enter your MFA code (or \'q\' to abort): '
        mfa_code = prompt_4_value(prompt_string, no_confirm = True, input = input)
        try:
            if mfa_code == 'q':
                return mfa_code
            int(mfa_code)
            mfa_code[5]
            break
        except:
            printError('Error: your MFA code must only consist of digits and be at least 6 characters long.')
    return mfa_code


def prompt_4_mfa_serial(input = None):
    """
    Prompt for an MFA serial number

    :param input:                       Used for unit testing

    :return:                            The MFA serial number
    """
    return prompt_4_value('Enter your MFA serial:', required = False, regex = re_mfa_serial_format, regex_format = mfa_serial_format, input = input)


def prompt_4_overwrite(filename, force_write, input = None):
    """
    Prompt whether the file should be overwritten

    :param filename:                    Name of the file about to be written
    :param force_write:                 Skip confirmation prompt if this flag is set
    :param input:                       Used for unit testing

    :return:                            Boolean whether file write operation is allowed
    """
    if not os.path.exists(filename) or force_write:
        return True
    return prompt_4_yes_no('File \'{}\' already exists. Do you want to overwrite it'.format(filename), input = input)


def prompt_4_value(question, choices = None, default = None, display_choices = True, display_indices = False, authorize_list = False, is_question = False, no_confirm = False, required = True, regex = None, regex_format = '', max_laps = 5, input = None):
    """
    Prompt for a value
                                        .                    .
    :param question:                    Question to be asked
    :param choices:                     List of authorized answers
    :param default:                     Value suggested by default
    :param display_choices:             Display accepted choices
    :param display_indices:             Display the indice in the list next to the choice
    :param authorize_list:              Set to true if a list of answers may be accepted
    :param is_question:                 Set to true to append a question mark
    :param no_confirm:                  Set to true to not prompt for a confirmation of the value
    :param required:                    Set to false if an empty answer is authorized
    :param regex:                       TODO
    :param regex_format                 TODO
    :param max_laps:                    Exit after N laps
    :param input:                       Used for unit testing

    :return:
    """
    if choices and display_choices and not display_indices:
        question = question + ' (' + '/'.join(choices) + ')'
    lap_n = 0
    while True:
        if lap_n >= max_laps:
            printError('Automatically abording prompt loop after 5 failures')
            return None
        lap_n += 1
        can_return = False
        # Display the question, choices, and prompt for the answer
        if is_question:
            question = question + '? '
        printError(question)
        if choices and display_indices:
            for c in choices:
                printError('%3d. %s' % (choices.index(c), c))
            printError('Enter the number corresponding to your choice: ', False)
        choice = prompt(input)
        # Set the default value if empty choice
        if not choice or choice == '':
            if default:
                if no_confirm or prompt_4_yes_no('Use the default value (' + default + ')'):
                    #return default
                    choice = default
                    can_return = True
            elif not required:
                can_return = True
            else:
                printError('Error: you cannot leave this parameter empty.')
        # Validate the value against a whitelist of choices
        elif choices:
            user_choices = [item.strip() for item in choice.split(',')]
            if not authorize_list and len(user_choices) > 1:
                printError('Error: multiple values are not supported; please enter a single value.')
            else:
                choice_valid = True
                if display_indices and int(choice) < len(choices):
                    choice = choices[int(choice)]
                else:
                    for c in user_choices:
                        if not c in choices:
                            printError('Invalid value (%s).' % c)
                            choice_valid = False
                            break
                if choice_valid:
                    can_return = True
        # Validate against a regex
        elif regex:
            if regex.match(choice):
                #return choice
                can_return = True
            else:
                printError('Error: expected format is: %s' % regex_format)
        else:
            # No automated validation, can attempt to return
            can_return = True
        if can_return:
            # Manually onfirm that the entered value is correct if needed
            if no_confirm or prompt_4_yes_no('You entered "' + choice + '". Is that correct', input=input):
                return choice


def prompt_4_yes_no(question, input = None):
    """
    Prompt for a yes/no or y/n answer
                                        .
    :param question:                    Question to be asked
    :param input:                       Used for unit testing

    :return:                            True for yes/y, False for no/n
    """
    count = 0
    while True:
        printError(question + ' (y/n)? ')
        choice = prompt(input).lower()
        if choice == 'yes' or choice == 'y':
            return True
        elif choice == 'no' or choice == 'n':
            return False
        else:
            count += 1
            printError('\'%s\' is not a valid answer. Enter \'yes\'(y) or \'no\'(n).' % choice)
            if count > 3:
                return None



# -*- coding: utf-8 -*-

import boto3
from botocore.session import Session
from collections import Counter

from opinel.utils.console import printInfo, printException



def build_region_list(service, chosen_regions = [], partition_name = 'aws'):
    """
    Build the list of target region names

    :param service:
    :param chosen_regions:
    :param partition_name:

    :return:
    """
    service = 'ec2containerservice' if service == 'ecs' else service # Of course things aren't that easy...
    # Get list of regions from botocore
    regions = Session().get_available_regions(service, partition_name = partition_name)
    if len(chosen_regions):
        return list((Counter(regions) & Counter(chosen_regions)).elements())
    else:
        return regions


def connect_service(service, credentials, region_name = None, config = None, silent = False):
    """
    Instantiates an AWS API client

    :param service:
    :param credentials:
    :param region_name:
    :param config:
    :param silent:

    :return:
    """
    api_client = None
    try:
        client_params = {}
        client_params['service_name'] = service.lower()
        session_params = {}
        session_params['aws_access_key_id'] = credentials['AccessKeyId']
        session_params['aws_secret_access_key'] = credentials['SecretAccessKey']
        session_params['aws_session_token'] = credentials['SessionToken']
        if region_name:
            client_params['region_name'] = region_name
            session_params['region_name'] = region_name
        if config:
            client_params['config'] = config
        aws_session = boto3.session.Session(**session_params)
        if not silent:
            infoMessage = 'Connecting to AWS %s' % service
            if region_name:
                infoMessage = infoMessage + ' in %s' % region_name
            printInfo('%s...' % infoMessage)
        api_client = aws_session.client(**client_params)
    except Exception as e:
        printException(e)
    return api_client


def get_name(src, dst, default_attribute):
    """

    :param src:
    :param dst:
    :param default_attribute:

    :return:
    """
    name_found = False
    if 'Tags' in src:
        for tag in src['Tags']:
            if tag['Key'] == 'Name' and tag['Value'] != '':
                dst['name'] = tag['Value']
                name_found = True
    if not name_found:
        dst['name'] = src[default_attribute]
    return dst['name']


def get_caller_identity(credentials):
    api_client = connect_service('sts', credentials, silent = True)
    return api_client.get_caller_identity()


def get_username(credentials):
    caller_identity = get_caller_identity(credentials)
    return caller_identity['Arn'].split('/')[-1]


def get_aws_account_id(credentials):
    caller_identity = get_caller_identity(credentials)
    return caller_identity['Arn'].split(':')[4]


def handle_truncated_response(callback, params, entities):
    """
    Handle truncated responses

    :param callback:
    :param params:
    :param entities:

    :return:
    """
    results = {}
    for entity in entities:
        results[entity] = []
    while True:
        marker_found = False
        response = callback(**params)
        for entity in entities:
            if entity in response:
                results[entity] = results[entity] + response[entity]
        for marker_name in ['NextToken', 'Marker']:
            if marker_name in response and response[marker_name]:
                params[marker_name] = response[marker_name]
                marker_found = True
        if not marker_found:
            break
    return results



#                            -*- coding: utf-8 -*-
                            
import boto3
from distutils.version import StrictVersion
import os
import re

from opinel import __version__ as OPINEL_VERSION
from opinel.utils.console import printError


########################################
# Regex
########################################

re_opinel = re.compile(r'^opinel>=([0-9.]+),<([0-9.]+).*')
re_boto3 = re.compile(r'^boto3>=([0-9.]+)(,<([0-9.]+).*)?')


########################################
# Functions
########################################

def check_requirements(script_path, requirements_file = None):
    """
    Check versions of opinel and boto3
    :param script_path:
    :return:
    """
    script_dir = os.path.dirname(script_path)
    opinel_min_version = opinel_max_version = boto3_min_version = boto3_max_version = None
    # Requirements file is either next to the script or in data/requirements
    if not requirements_file:
        requirements_file = os.path.join(script_dir, 'data/requirements.txt') if os.path.isfile(os.path.join(script_dir, 'data/requirements.txt')) else 'requirements.txt'
    with open(requirements_file, 'rt') as f:
        for requirement in f.readlines():
            opinel_requirements = re_opinel.match(requirement)
            if opinel_requirements:
                opinel_requirements = opinel_requirements.groups()
                opinel_min_version = opinel_requirements[0]
                opinel_max_version = opinel_requirements[1]
            boto3_requirements = re_boto3.match(requirement)
            if boto3_requirements:
                boto3_requirements = boto3_requirements.groups()
                boto3_min_version = boto3_requirements[0]
                boto3_max_version = boto3_requirements[1]
    if not check_versions(opinel_min_version, OPINEL_VERSION, opinel_max_version, 'opinel'):
        return False
    if not check_versions(boto3_min_version, boto3.__version__, boto3_max_version, 'boto3'):
        return False
    return True


def check_versions(min_version, installed_version, max_version, package_name, strict = False):
    """

    :param min_version:
    :param installed_version:
    :param max_version:
    :param package_name:

    :return:
    """
    if not min_version:
        # If no minimum version was specified, pass
        return True
    if StrictVersion(installed_version) < StrictVersion(min_version):
        printError('Error: the version of %s installed on this system (%s) is too old. You need at least version %s to run this tool.' % (package_name, OPINEL_VERSION, min_version))
        return False
    if max_version and StrictVersion(installed_version) >= StrictVersion(max_version):
        printError('Warning: ther version of %s installed on this system (%s) is too recent; you may experience unexpected runtime errors as versions above %s have not been tested.' % (package_name, installed_version, max_version))
        if strict:
            printError('Warning treated as error.')
            return False
    return True


def manage_dictionary(dictionary, key, init, callback = None):
    """

    :param dictionary:
    :param key:
    :param init:
    :param callback:

    :return:
    """
    if not str(key) in dictionary:
        dictionary[str(key)] = init
        manage_dictionary(dictionary, key, init)
        if callback:
            callback(dictionary[key])
    return dictionary



# -*- coding: utf-8 -*-
from __future__ import print_function

import datetime
import json
import os
import re
import sys

from opinel.utils.console import printException, prompt_4_overwrite
from opinel.utils.conditions import pass_condition



class CustomJSONEncoder(json.JSONEncoder):
    """
    JSON encoder class
    """
    def default(self, o):
        if type(o) == datetime.datetime:
            return str(o)
        else:
            return o.__dict__


def load_data(data_file, key_name = None, local_file = False):
    """
    Load a JSON data file

    :param data_file:
    :param key_name:
    :param local_file:
    :return:
    """
    if local_file:
        if data_file.startswith('/'):
            src_file = data_file
        else:
            src_dir = os.getcwd()
            src_file = os.path.join(src_dir, data_file)
    else:
        src_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../data')
        src_file = os.path.join(src_dir, data_file)
    with open(src_file) as f:
        data = json.load(f)
    if key_name:
        data = data[key_name]
    return data


def read_ip_ranges(filename, local_file = True, conditions = [], ip_only = False):
    """
    Returns the list of IP prefixes from an ip-ranges file

    :param filename:
    :param local_file:
    :param conditions:
    :param ip_only:
    :return:
    """
    targets = []
    data = load_data(filename, 'prefixes', local_file)
    for d in data:
        condition_passed = True
        for condition in conditions:
            condition_passed = pass_condition(d[condition[0]], condition[1], condition[2])
            if not condition_passed:
                break
        if condition_passed:
            targets.append(d)
    if ip_only:
        ips = []
        for t in targets:
            ips.append(t['ip_prefix'])
        return ips
    else:
        return targets


def save_blob_as_json(filename, blob, force_write, debug):
    """
    Creates/Modifies file and saves python object as JSON

    :param filename:
    :param blob:
    :param force_write:
    :param debug:

    :return:
    """
    try:
        if prompt_4_overwrite(filename, force_write):
            with open(filename, 'wt') as f:
                print('%s' % json.dumps(blob, indent=4 if debug else None, separators=(',', ': '), sort_keys=True, cls=CustomJSONEncoder), file=f)
    except Exception as e:
        printException(e)
        pass


def save_ip_ranges(profile_name, prefixes, force_write, debug):
    """
    Creates/Modifies an ip-range-XXX.json file

    :param profile_name:
    :param prefixes:
    :param force_write:
    :param debug:

    :return:
    """
    filename = 'ip-ranges-%s.json' % profile_name
    ip_ranges = {}
    ip_ranges['createDate'] = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
    ip_ranges['prefixes'] = prefixes
    save_blob_as_json(filename, ip_ranges, force_write, debug)



# -*- coding: utf-8 -*-

from threading import Thread
try:
    # Python2
    from Queue import Queue
except ImportError:
    # Python3
    from queue import Queue

from opinel.utils.console import printException



def thread_work(targets, function, params = {}, num_threads = 0):
    """
    Generic multithreading helper

    :param targets:
    :param function:
    :param params:
    :param num_threads:

    :return:
    """
    q = Queue(maxsize=0)
    if not num_threads:
        num_threads = len(targets)
    for i in range(num_threads):
        worker = Thread(target=function, args=(q, params))
        worker.setDaemon(True)
        worker.start()
    for target in targets:
        q.put(target)
    q.join()


def threaded_per_region(q, params):
    """
    Helper for multithreading on a per-region basis

    :param q:
    :param params:

    :return:
    """
    while True:
        try:
            params['region'] = q.get()
            method = params['method']
            method(params)
        except Exception as e:
            printException(e)
        finally:
            q.task_done()



# -*- coding: utf-8 -*-


def get_s3_bucket_location(s3_client, bucket_name):
    """

    :param s3_client:
    :param bucket_name:
    :return:
    """
    location = s3_client.get_bucket_location(Bucket = bucket_name)
    return location['LocationConstraint'] if location['LocationConstraint'] else 'us-east-1'



# -*- coding: utf-8 -*-

import argparse
import json
import os
import sys

opinel_arg_dir = os.path.join(os.path.expanduser('~'), '.aws/opinel')

class OpinelArgumentParser(object):
    """
    """

    def __init__(self, tool_name = ''):
        self.parser = argparse.ArgumentParser()
        self.default_args = read_default_args(tool_name)

    def add_argument(self, argument_name, help = None, dest = None, nargs = None, default = None, action = None, choices = None):

        # Built-in, common arguments
        if argument_name == 'debug':
            self.parser.add_argument('--debug',
                                dest='debug',
                                default=False,
                                action='store_true',
                                help='Print the stack trace when exception occurs' if not help else help)
        elif argument_name == 'dry-run':
            self.parser.add_argument('--dry-run',
                                dest='dry_run',
                                default=False,
                                action='store_true',
                                help='Executes read-only actions (check status, describe*, get*, list*...)' if not help else help)
        elif argument_name == 'profile':
            self.parser.add_argument('--profile',
                                dest='profile',
                                default= ['default'],
                                nargs='+',
                                help='Name of the profile' if not help else help)
        elif argument_name == 'regions':
            self.parser.add_argument('--regions',
                                dest='regions',
                                default=[],
                                nargs='+',
                                help='Name of regions to run the tool in, defaults to all' if not help else help)
        elif argument_name == 'partition-name':
            self.parser.add_argument('--partition-name',
                                dest='partition_name',
                                default='aws',
                                help='Switch out of the public AWS partition (e.g. US gov or China)')
        elif argument_name == 'vpc':
            self.parser.add_argument('--vpc',
                                dest='vpc',
                                default=[],
                                nargs='+',
                                help='Name of VPC to run the tool in, defaults to all' if not help else help)
        elif argument_name == 'force':
            self.parser.add_argument('--force',
                                dest='force_write',
                                default=False,
                                action='store_true',
                                help='Overwrite existing files' if not help else help)
        elif argument_name == 'ip-ranges':
            self.parser.add_argument('--ip-ranges',
                                dest='ip_ranges',
                                default=[],
                                nargs='+',
                                help='Config file(s) that contain your known IP ranges.' if not help else help)
        elif argument_name == 'ip-ranges-name-key':
            self.parser.add_argument('--ip-ranges-name-key',
                                dest='ip_ranges_name_key',
                                default='name',
                                help='Name of the key containing the display name of a known CIDR.' if not help else help)
        elif argument_name == 'mfa-serial':
            self.parser.add_argument('--mfa-serial',
                                dest='mfa_serial',
                                default=None,
                                help='ARN of the user\'s MFA device' if not help else help)
        elif argument_name == 'mfa-code':
            self.parser.add_argument('--mfa-code',
                                dest='mfa_code',
                                default=None,
                                help='Six-digit code displayed on the MFA device.' if not help else help)
        elif argument_name == 'csv-credentials':
            self.parser.add_argument('--csv-credentials',
                                dest='csv_credentials',
                                default=None,
                                help='Path to a CSV file containing the access key ID and secret key' if not help else help)
        elif argument_name == 'user-name':
            self.parser.add_argument('--user-name',
                                dest='user_name',
                                default=[],
                                nargs='+',
                                help='Name of the user.' if not help else help)
        elif argument_name == 'bucket-name':
            self.parser.add_argument('--bucket-name',
                                dest='bucket_name',
                                default=[None],
                                help='Name of the s3 bucket.' if not help else help)
        elif argument_name == 'group-name':
            self.parser.add_argument('--group-name',
                                dest='group_name',
                                default=[],
                                nargs='+',
                                help='Name of the IAM group.' if not help else help)

        # Default
        elif help != None and default != None and (nargs != None or action != None):
            dest = argument_name.replace('-', '_') if not dest else dest
            if nargs:
                if not choices:
                    self.parser.add_argument('--%s' % argument_name,
                                             dest = dest,
                                             default = self.default_args[dest] if dest in self.default_args else default,
                                             nargs = nargs,
                                             help = help)
                else:
                    self.parser.add_argument('--%s' % argument_name,
                                             dest = dest,
                                             default = self.default_args[dest] if dest in self.default_args else default,
                                             nargs = nargs,
                                             choices = choices,
                                             help = help)
            elif action:
                self.parser.add_argument('--%s' % argument_name,
                                         dest = dest,
                                         default = self.default_args[dest] if dest in self.default_args else default,
                                         action = action,
                                         help = help)

        # Error
        else:
            raise Exception('Invalid parameter name %s' % argument_name)

    
    def parse_args(self):
        args = self.parser.parse_args()
        return args


def read_default_args(tool_name):
    """
    Read default argument values for a given tool

    :param tool_name:                   Name of the script to read the default arguments for
    :return:                            Dictionary of default arguments (shared + tool-specific)
    """
    profile_name = 'default'
    # h4ck to have an early read of the profile name
    for i, arg in enumerate(sys.argv):
        if arg == '--profile' and len(sys.argv) >= i + 1:
            profile_name = sys.argv[i + 1]
    if not os.path.isdir(opinel_arg_dir):
        os.makedirs(opinel_arg_dir)
    opinel_arg_file = os.path.join(opinel_arg_dir, '%s.json' % profile_name)
    default_args = {}
    if os.path.isfile(opinel_arg_file):
        with open(opinel_arg_file, 'rt') as f:
            all_args = json.load(f)
        for target in all_args:
            if tool_name.endswith(target):
                default_args.update(all_args[target])
        for k in all_args['shared']:
            if k not in default_args:
                default_args[k] = all_args['shared'][k]
    return default_args



#!/usr/bin/env python

# distutils/setuptools install script for opinel
import os
from setuptools import setup, find_packages

# Package info
NAME = 'opinel'
ROOT = os.path.dirname(__file__)
VERSION = __import__(NAME).__version__

# Requirements
requirements = []
with open('requirements.txt') as f:
    for r in f.readlines():
        requirements.append(r.strip())

# Setup
setup(
    name=NAME,
    version=VERSION,
    description='Code shared between Scout2 and AWS-recipes.',
    long_description=open('README.rst').read(),
    author='l01cd3v',
    author_email='l01cd3v@gmail.com',
    url='https://github.com/nccgroup/opinel',
    packages=[
        'opinel', 'opinel.utils', 'opinel.services'
    ],
    package_data={
        'opinel': [
            'data/*.json',
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license='GNU General Public License v2 (GPLv2)',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4'
    ],
)
